#!/usr/bin/env python3
import os
import sys
import multiprocessing as mp
from itertools import repeat
import pandas
import regex
from Bio import SeqIO


def compl_DNA(c) -> str:
	'''Return complemetary nucleotide

	In this script only uppercase nucleotides may appear, otherwise:
	if c.islower():
		c = c.upper()
	'''
	if c == 'A':
		return 'T'
	if c == 'C':
		return 'G'
	if c == 'G':
		return 'C'
	if c == 'T':
		return 'A'
	return c


def rev_compl_DNA(s) -> str:
	'''Return reverse complement DNA'''
	return ''.join(list(map(compl_DNA, s[::-1])))


def make_kmer_list(k_len=4, nr=True):
	'''Return list of kmers different (if nr) from their reverse complements'''
	acgt = ['A', 'C', 'G', 'T']
	kmers = acgt
	for _i in range(k_len - 1):
		kmers = [f"{x}{y}" for x in kmers for y in acgt]
	if nr is False:
		return kmers
	nr_kmers = []
	for x in kmers:
		if rev_compl_DNA(x) not in nr_kmers:
			nr_kmers.append(x)
	return nr_kmers


def kmers_freq_single(record, patterns, klen):
	'''Return record.id, record.len and kmer frequences multiplied by 1000.'''
	d = {}
	seq_len = len(record.seq)
	seq = str(record.seq).upper()
	for (i, j) in patterns:
		d[i] = len(regex.findall(j, seq, overlapped=True))
	seq_rc = str(record.seq.reverse_complement()).upper()
	for (i, j) in patterns:
		d[i] = d.get(i, 0) + len(regex.findall(j, seq_rc, overlapped=True))
	freq_list = [1000 * d.get(i, 0) / (2 * (seq_len - klen)) for i, _ in patterns]
	return [record.id, seq_len] + freq_list


def kmers_freq(records, kmers, num_pools=1):
	patterns = [(i, regex.compile(i)) for i in kmers]
	klen = len(kmers[0])
	with mp.Pool(num_pools) as p:
		return p.starmap(
			kmers_freq_single,
			zip(records, repeat(patterns), repeat(klen)))


def kmer_freq_table(filename, k_len=4, num_pools=1):
	kmer_list = make_kmer_list(k_len, nr=True)
	return pandas.DataFrame.from_records(
		kmers_freq(SeqIO.parse(filename, "fasta"), kmer_list, num_pools),
		columns=['fragment', 'length'] + kmer_list)


if __name__ == '__main__':
	outfilename = f"tn.{os.path.basename(sys.argv[1])}.csv"
	kmer_freq_table(sys.argv[1]).to_csv(outfilename)



#!/usr/bin/env python3
import logging
import os
import os.path
from pyyamb.utils import run_external
import pysam


def map_reads(args):
	'''Return list of alignments'''
	sams = []
	if args.single_end:
		for read in args.single_end:
			sams.append(call_minimap2(args, [read]))
	if args.pe_1 and args.pe_1:
		# Only one pair of reads is currently mapped
		reads = list(zip(args.pe_1, args.pe_2))
		for pair in reads:
			sams.append(call_minimap2(args, pair))
	return sams


def call_minimap2(args, reads):
	logger = logging.getLogger("main")
	prefix, ext = os.path.splitext(os.path.basename(reads[0]))
	if ext in ['gz', 'bz2']:
		prefix, _ = os.path.splitext(prefix)
	target = os.path.join(args.output, f"{prefix}.sam")
	cmd = [
		"minimap2",
		"-x", "sr",
		"-t", str(args.threads),
		"-a",
		"-o", target,
		args.assembly, *reads]

	try:
		logger.info("Mapping reads: %s", ", ".join(reads))
		run_external(cmd)
		return target
	except Exception as e:
		logger.error("Unsuccesful mapping.")
		raise e


def view_mapping_file(args, mapping_sam_file, compress=False):
	logger = logging.getLogger("main")
	logger.info("Converting mapping file")
	prefix, _ = os.path.splitext(mapping_sam_file)
	mapping_bam_file = os.path.join(args.output, f'{prefix}.bam')
	opts = [
		'-@', str(args.threads),
		'-F', '0x4',
		'-o', mapping_bam_file,
		mapping_sam_file]
	if not compress:
		opts.insert(0, '-u')
	pysam.view(*opts, catch_stdout=False)
	os.remove(mapping_sam_file)

	return mapping_bam_file


def sort_mapping_file(args, mapping_bam_file):
	logger = logging.getLogger("main")
	logger.info("Sorting mapping file")
	prefix, _ = os.path.splitext(mapping_bam_file)
	sorted_mapping_bam_file = os.path.join(args.output, f'{prefix}.sorted.bam')
	'''User-provided memory limit not set here cause it's a memory per thread
	'-m', f'{args.memory_limit}G'
	'''
	pysam.sort(
		'-@', str(args.threads),
		'-m', '1G',
		'-o', sorted_mapping_bam_file,
		mapping_bam_file)
	os.remove(mapping_bam_file)
	logger.info("Indexing mapping file")
	pysam.samtools.index(sorted_mapping_bam_file)

	return sorted_mapping_bam_file



import logging
import os.path
import subprocess
from Bio import SeqIO
import regex


def run_external(cmd, keep_stdout=False, keep_stderr=True):
	logger = logging.getLogger("main")
	logger.debug("Running: %s", " ".join(cmd))
	stderr_dest = subprocess.PIPE if keep_stderr else subprocess.DEVNULL
	stdout_dest = subprocess.PIPE if keep_stdout else subprocess.DEVNULL

	try:
		r = subprocess.run(cmd, stderr=stderr_dest, stdout=stdout_dest,
			check=True, encoding="utf-8")
		return r
	except subprocess.CalledProcessError as e:
		logger.error("Error during run of \"%s\"", e.cmd)
		logger.error("stderr message:")
		logger.error(e.stderr)
		raise e


def check_files(file_list):
	for x in file_list:
		if not os.path.isfile(x):
			logging.error("Input file not found: %s", x)
			raise FileNotFoundError(x)


def write_records_to_fasta(records, path, glue=False):
	'''Return path to FASTA or None'''
	logger = logging.getLogger()
	with open(path, "w") as f:
		if not glue:
			for record in records:
				SeqIO.write(record, f, 'fasta')
		else:
			pattern = regex.compile(r'^(.+)_frag_(\d+)$')
			prev_rec = next(records)
			m0 = pattern.match(prev_rec.id)
			if m0:
				prev_cont = m0.group(1)
				prev_n = int(m0.group(2))
			else:
				prev_cont = prev_rec.id
				prev_n = -127
			for record in records:
				m = pattern.match(record.id)
				if m:
					contig = m.group(1)
					n = int(m.group(2))
					if contig != prev_cont:
						prev_rec.id = f"{prev_cont}_{prev_n}"
						prev_rec.description = ''
						SeqIO.write(prev_rec, f, 'fasta')
						prev_cont = contig
						prev_rec = record
					else:
						if n == prev_n + 1:
							prev_rec += record
						else:
							prev_rec.id = f"{prev_cont}_{prev_n}"
							prev_rec.description = ''
							SeqIO.write(prev_rec, f, 'fasta')
							prev_cont = contig
							prev_rec = record
					prev_n = n
				else:
					prev_rec.id = prev_cont
					prev_rec.description = ''
					SeqIO.write(prev_rec, f, 'fasta')
					prev_cont = record.id
					prev_rec = record
			SeqIO.write(prev_rec, f, 'fasta')

		logger.debug("Sequences had been written to: %s", path)
		return path
	return None



import pyyamb
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyyamb",
    version=pyyamb.__version__,
    author="Aleksei Korzhenkov",
    author_email="oscypek@ya.ru",
    description="Python implementation of YAMB (Yet Another Metagenome Binner)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/laxeye/pyyamb",
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pyyamb = pyyamb.pyyamb:main',
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    python_requires='>=3.10',
    install_requires=[
        'biopython',
        'pandas',
        'seaborn',
        'pysam',
        'matplotlib',
        'numpy',
        'scikit-learn>=1.3',
        'pycoverm',
        'regex'
    ]
)
